#ifndef FINESSE_CONSTANTS_H_
#define FINESSE_CONSTANTS_H_

#include <complex.h>
#include <float.h>

double C_LIGHT = 299792458.0;
double H_PLANCK = 6.62607015e-34;
long double PI = 3.14159265358979323846;
double TWO_ON_PI_QRT = 0.8932438417380023;
double ROOT2 = 1.4142135623730951;
double INV_ROOT2 = 0.7071067811865475;
double DEG2RAD = 3.14159265358979323846/180;
double RAD2DEG = 180/3.14159265358979323846;
double R2DB = 8.685889638065037;
double DB2R = 1/8.685889638065037;

double complex COMPLEX_0 = 0.0 + 0.0*I;
double complex COMPLEX_1 = 1.0 + 0.0*I;
double complex COMPLEX_I = 0.0 + 1.0*I;

#endif // !FINESSE_CONSTANTS_H_
