from finesse.element cimport BaseCValues, ElementWorkspace


cdef class VariableValues(BaseCValues):
    cdef public:
        double value


cdef class VariableWorkspace(ElementWorkspace):
    cdef:
        VariableValues vv
